@@ -67,12 +67,6 @@ static struct
 } entry_features;
 
 
-/* The control variable of the connection owning the current pinentry.
-   This is only valid if ENTRY_CTX is not NULL.  Note, that we care
-   only about the value of the pointer and that it should never be
-   dereferenced.  */
-static ctrl_t entry_owner;
-
 /* A mutex used to serialize access to the pinentry. */
 static npth_mutex_t entry_lock;
 
@@ -104,11 +98,15 @@ void
 initialize_module_call_pinentry (void)
 {
   static int initialized;
+  int err;
 
   if (!initialized)
     {
-      if (npth_mutex_init (&entry_lock, NULL))
-        initialized = 1;
+      err = npth_mutex_init (&entry_lock, NULL);
+      if (err)
+	log_fatal ("error initializing mutex: %s\n", strerror (err));
+
+      initialized = 1;
     }
 }
 
@@ -128,7 +126,7 @@ agent_query_dump_state (void)
 void
 agent_reset_query (ctrl_t ctrl)
 {
-  if (entry_ctx && popup_tid && entry_owner == ctrl)
+  if (entry_ctx && popup_tid && ctrl->pinentry_active)
     {
       agent_popup_message_stop (ctrl);
     }
@@ -140,7 +138,7 @@ agent_reset_query (ctrl_t ctrl)
    stalled pinentry does not block other threads.  Fixme: We should
    have a timeout in Assuan for the disconnect operation. */
 static gpg_error_t
-unlock_pinentry (gpg_error_t rc)
+unlock_pinentry (ctrl_t ctrl, gpg_error_t rc)
 {
   assuan_context_t ctx = entry_ctx;
   int err;
@@ -177,19 +175,53 @@ unlock_pinentry (gpg_error_t rc)
         }
     }
 
-  entry_ctx = NULL;
-  err = npth_mutex_unlock (&entry_lock);
-  if (err)
+  if (--ctrl->pinentry_active == 0)
     {
-      log_error ("failed to release the entry lock: %s\n", strerror (err));
-      if (!rc)
-        rc = gpg_error_from_errno (err);
+      entry_ctx = NULL;
+      err = npth_mutex_unlock (&entry_lock);
+      if (err)
+        {
+          log_error ("failed to release the entry lock: %s\n", strerror (err));
+          if (!rc)
+            rc = gpg_error_from_errno (err);
+        }
+      assuan_release (ctx);
     }
-  assuan_release (ctx);
   return rc;
 }
 
 
+/* Helper for at_fork_cb which can also be called by the parent to
+ * show shich envvars will be set.  */
+static void
+atfork_core (ctrl_t ctrl, int debug_mode)
+{
+  int iterator = 0;
+  const char *name, *assname, *value;
+
+  while ((name = session_env_list_stdenvnames (&iterator, &assname)))
+    {
+      /* For all new envvars (!ASSNAME) and the two medium old ones
+       * which do have an assuan name but are conveyed using
+       * environment variables, update the environment of the forked
+       * process.  */
+      if (!assname
+          || !strcmp (name, "XAUTHORITY")
+          || !strcmp (name, "PINENTRY_USER_DATA"))
+        {
+          value = session_env_getenv (ctrl->session_env, name);
+          if (value)
+            {
+              if (debug_mode)
+                log_debug ("pinentry: atfork used setenv(%s,%s)\n",name,value);
+              else
+                gnupg_setenv (name, value, 1);
+            }
+        }
+    }
+}
+
+
 /* To make sure we leave no secrets in our image after forking of the
    pinentry, we use this callback. */
 static void
@@ -199,26 +231,8 @@ atfork_cb (void *opaque, int where)
 
   if (!where)
     {
-      int iterator = 0;
-      const char *name, *assname, *value;
-
       gcry_control (GCRYCTL_TERM_SECMEM);
-
-      while ((name = session_env_list_stdenvnames (&iterator, &assname)))
-        {
-          /* For all new envvars (!ASSNAME) and the two medium old
-             ones which do have an assuan name but are conveyed using
-             environment variables, update the environment of the
-             forked process.  */
-          if (!assname
-              || !strcmp (name, "XAUTHORITY")
-              || !strcmp (name, "PINENTRY_USER_DATA"))
-            {
-              value = session_env_getenv (ctrl->session_env, name);
-              if (value)
-                gnupg_setenv (name, value, 1);
-            }
-        }
+      atfork_core (ctrl, 0);
     }
 }
 
@@ -288,6 +302,14 @@ start_pinentry (ctrl_t ctrl)
   char *flavor_version;
   int err;
 
+  if (ctrl->pinentry_active)
+    {
+      /* It's trying to use pinentry recursively.  In this situation,
+         the thread holds ENTRY_LOCK already.  */
+      ctrl->pinentry_active++;
+      return 0;
+    }
+
   npth_clock_gettime (&abstime);
   abstime.tv_sec += LOCK_TIMEOUT;
   err = npth_mutex_timedlock (&entry_lock, &abstime);
@@ -302,8 +324,6 @@ start_pinentry (ctrl_t ctrl)
       return rc;
     }
 
-  entry_owner = ctrl;
-
   if (entry_ctx)
     return 0;
 
@@ -325,7 +345,7 @@ start_pinentry (ctrl_t ctrl)
          the Wine implementation does not flush stdin,stdout and stderr
          - see above.  Let's try to ignore the error. */
 #ifndef HAVE_W32_SYSTEM
-      return unlock_pinentry (tmperr);
+      return unlock_pinentry (ctrl, tmperr);
 #endif
     }
 
@@ -371,6 +391,10 @@ start_pinentry (ctrl_t ctrl)
       log_error ("can't allocate assuan context: %s\n", gpg_strerror (rc));
       return rc;
     }
+
+  ctrl->pinentry_active = 1;
+  entry_ctx = ctx;
+
   /* We don't want to log the pinentry communication to make the logs
      easier to read.  We might want to add a new debug option to enable
      pinentry logging.  */
@@ -382,85 +406,86 @@ start_pinentry (ctrl_t ctrl)
      that atfork is used to change the environment for pinentry.  We
      start the server in detached mode to suppress the console window
      under Windows.  */
-  rc = assuan_pipe_connect (ctx, full_pgmname, argv,
+  rc = assuan_pipe_connect (entry_ctx, full_pgmname, argv,
 			    no_close_list, atfork_cb, ctrl,
 			    ASSUAN_PIPE_CONNECT_DETACHED);
   if (rc)
     {
       log_error ("can't connect to the PIN entry module '%s': %s\n",
                  full_pgmname, gpg_strerror (rc));
-      assuan_release (ctx);
-      return unlock_pinentry (gpg_error (GPG_ERR_NO_PIN_ENTRY));
+      return unlock_pinentry (ctrl, gpg_error (GPG_ERR_NO_PIN_ENTRY));
     }
-  entry_ctx = ctx;
 
   if (DBG_IPC)
     log_debug ("connection to PIN entry established\n");
 
+  if (opt.debug_pinentry)
+    atfork_core (ctrl, 1);
+
   value = session_env_getenv (ctrl->session_env, "PINENTRY_USER_DATA");
   if (value != NULL)
     {
       char *optstr;
       if (asprintf (&optstr, "OPTION pinentry-user-data=%s", value) < 0 )
-	return unlock_pinentry (out_of_core ());
+        return unlock_pinentry (ctrl, out_of_core ());
       rc = assuan_transact (entry_ctx, optstr, NULL, NULL, NULL, NULL, NULL,
 			    NULL);
       xfree (optstr);
       if (rc && gpg_err_code (rc) != GPG_ERR_UNKNOWN_OPTION)
-        return unlock_pinentry (rc);
+        return unlock_pinentry (ctrl, rc);
     }
 
   rc = assuan_transact (entry_ctx,
                         opt.no_grab? "OPTION no-grab":"OPTION grab",
                         NULL, NULL, NULL, NULL, NULL, NULL);
   if (rc)
-    return unlock_pinentry (rc);
+    return unlock_pinentry (ctrl, rc);
 
   value = session_env_getenv (ctrl->session_env, "GPG_TTY");
   if (value)
     {
       char *optstr;
       if (asprintf (&optstr, "OPTION ttyname=%s", value) < 0 )
-	return unlock_pinentry (out_of_core ());
+        return unlock_pinentry (ctrl, out_of_core ());
       rc = assuan_transact (entry_ctx, optstr, NULL, NULL, NULL, NULL, NULL,
 			    NULL);
       xfree (optstr);
       if (rc)
-	return unlock_pinentry (rc);
+        return unlock_pinentry (ctrl, rc);
     }
   value = session_env_getenv (ctrl->session_env, "TERM");
   if (value)
     {
       char *optstr;
       if (asprintf (&optstr, "OPTION ttytype=%s", value) < 0 )
-	return unlock_pinentry (out_of_core ());
+        return unlock_pinentry (ctrl, out_of_core ());
       rc = assuan_transact (entry_ctx, optstr, NULL, NULL, NULL, NULL, NULL,
 			    NULL);
       xfree (optstr);
       if (rc)
-	return unlock_pinentry (rc);
+        return unlock_pinentry (ctrl, rc);
     }
   if (ctrl->lc_ctype)
     {
       char *optstr;
       if (asprintf (&optstr, "OPTION lc-ctype=%s", ctrl->lc_ctype) < 0 )
-	return unlock_pinentry (out_of_core ());
+        return unlock_pinentry (ctrl, out_of_core ());
       rc = assuan_transact (entry_ctx, optstr, NULL, NULL, NULL, NULL, NULL,
 			    NULL);
       xfree (optstr);
       if (rc)
-	return unlock_pinentry (rc);
+        return unlock_pinentry (ctrl, rc);
     }
   if (ctrl->lc_messages)
     {
       char *optstr;
       if (asprintf (&optstr, "OPTION lc-messages=%s", ctrl->lc_messages) < 0 )
-	return unlock_pinentry (out_of_core ());
+        return unlock_pinentry (ctrl, out_of_core ());
       rc = assuan_transact (entry_ctx, optstr, NULL, NULL, NULL, NULL, NULL,
 			    NULL);
       xfree (optstr);
       if (rc)
-	return unlock_pinentry (rc);
+        return unlock_pinentry (ctrl, rc);
     }
 
 
@@ -476,7 +501,7 @@ start_pinentry (ctrl_t ctrl)
       rc = assuan_transact (entry_ctx, "OPTION allow-external-password-cache",
                             NULL, NULL, NULL, NULL, NULL, NULL);
       if (rc && gpg_err_code (rc) != GPG_ERR_UNKNOWN_OPTION)
-        return unlock_pinentry (rc);
+        return unlock_pinentry (ctrl, rc);
     }
 
   if (opt.allow_emacs_pinentry)
@@ -486,7 +511,7 @@ start_pinentry (ctrl_t ctrl)
       rc = assuan_transact (entry_ctx, "OPTION allow-emacs-prompt",
                             NULL, NULL, NULL, NULL, NULL, NULL);
       if (rc && gpg_err_code (rc) != GPG_ERR_UNKNOWN_OPTION)
-        return unlock_pinentry (rc);
+        return unlock_pinentry (ctrl, rc);
     }
 
 
@@ -524,7 +549,7 @@ start_pinentry (ctrl_t ctrl)
         if (*s == '|' && (s2=strchr (s+1,'|')))
           s = s2+1;
         if (asprintf (&optstr, "OPTION default-%s=%s", tbl[idx].key, s) < 0 )
-          return unlock_pinentry (out_of_core ());
+          return unlock_pinentry (ctrl, out_of_core ());
         assuan_transact (entry_ctx, optstr, NULL, NULL, NULL, NULL, NULL,
                          NULL);
         xfree (optstr);
@@ -607,7 +632,7 @@ start_pinentry (ctrl_t ctrl)
 
   /* Ask the pinentry for its version and flavor and store that as a
    * string in MB.  This information is useful for helping users to
-   * figure out Pinentry problems.  Noet that "flavor" may also return
+   * figure out Pinentry problems.  Note that "flavor" may also return
    * a status line with the features; we use a dedicated handler for
    * that.  */
   {
@@ -651,8 +676,8 @@ start_pinentry (ctrl_t ctrl)
       rc = agent_inq_pinentry_launched (ctrl, pinentry_pid, flavor_version);
       if (gpg_err_code (rc) == GPG_ERR_CANCELED
           || gpg_err_code (rc) == GPG_ERR_FULLY_CANCELED)
-        return unlock_pinentry (gpg_err_make (GPG_ERR_SOURCE_DEFAULT,
-                                              gpg_err_code (rc)));
+        return unlock_pinentry (ctrl, gpg_err_make (GPG_ERR_SOURCE_DEFAULT,
+                                                    gpg_err_code (rc)));
       rc = 0;
     }
 
@@ -882,13 +907,6 @@ setup_qualitybar (ctrl_t ctrl)
   return 0;
 }
 
-enum
-  {
-    PINENTRY_STATUS_CLOSE_BUTTON = 1 << 0,
-    PINENTRY_STATUS_PIN_REPEATED = 1 << 8,
-    PINENTRY_STATUS_PASSWORD_FROM_CACHE = 1 << 9
-  };
-
 /* Check the button_info line for a close action.  Also check for the
    PIN_REPEATED flag.  */
 static gpg_error_t
@@ -953,7 +971,6 @@ agent_askpin (ctrl_t ctrl,
   const char *errtext = NULL;
   int is_pin = 0;
   int saveflag;
-  unsigned int pinentry_status;
 
   if (opt.batch)
     return 0; /* fixme: we should return BAD PIN */
@@ -1022,18 +1039,18 @@ agent_askpin (ctrl_t ctrl,
   rc = assuan_transact (entry_ctx, line,
 			NULL, NULL, NULL, NULL, NULL, NULL);
   if (rc && gpg_err_code (rc) != GPG_ERR_ASS_UNKNOWN_CMD)
-    return unlock_pinentry (rc);
+    return unlock_pinentry (ctrl, rc);
 
   build_cmd_setdesc (line, DIM(line), desc_text);
   rc = assuan_transact (entry_ctx, line, NULL, NULL, NULL, NULL, NULL, NULL);
   if (rc)
-    return unlock_pinentry (rc);
+    return unlock_pinentry (ctrl, rc);
 
   snprintf (line, DIM(line), "SETPROMPT %s",
             prompt_text? prompt_text : is_pin? L_("PIN:") : L_("Passphrase:"));
   rc = assuan_transact (entry_ctx, line, NULL, NULL, NULL, NULL, NULL, NULL);
   if (rc)
-    return unlock_pinentry (rc);
+    return unlock_pinentry (ctrl, rc);
 
   /* If a passphrase quality indicator has been requested and a
      minimum passphrase length has not been disabled, send the command
@@ -1042,7 +1059,7 @@ agent_askpin (ctrl_t ctrl,
     {
       rc = setup_qualitybar (ctrl);
       if (rc)
-        return unlock_pinentry (rc);
+        return unlock_pinentry (ctrl, rc);
     }
 
   if (initial_errtext)
@@ -1051,7 +1068,7 @@ agent_askpin (ctrl_t ctrl,
       rc = assuan_transact (entry_ctx, line,
                             NULL, NULL, NULL, NULL, NULL, NULL);
       if (rc)
-        return unlock_pinentry (rc);
+        return unlock_pinentry (ctrl, rc);
     }
 
   if (pininfo->with_repeat)
@@ -1064,6 +1081,7 @@ agent_askpin (ctrl_t ctrl,
         pininfo->with_repeat = 0; /* Pinentry does not support it.  */
     }
   pininfo->repeat_okay = 0;
+  pininfo->status = 0;
 
   for (;pininfo->failed_tries < pininfo->max_tries; pininfo->failed_tries++)
     {
@@ -1076,13 +1094,14 @@ agent_askpin (ctrl_t ctrl,
         {
           /* TRANSLATORS: The string is appended to an error message in
              the pinentry.  The %s is the actual error message, the
-             two %d give the current and maximum number of tries. */
+             two %d give the current and maximum number of tries.
+             Do not translate the "SETERROR" keyword. */
           snprintf (line, DIM(line), L_("SETERROR %s (try %d of %d)"),
                     errtext, pininfo->failed_tries+1, pininfo->max_tries);
           rc = assuan_transact (entry_ctx, line,
                                 NULL, NULL, NULL, NULL, NULL, NULL);
           if (rc)
-            return unlock_pinentry (rc);
+            return unlock_pinentry (ctrl, rc);
           errtext = NULL;
         }
 
@@ -1092,15 +1111,14 @@ agent_askpin (ctrl_t ctrl,
           rc = assuan_transact (entry_ctx, line,
                                 NULL, NULL, NULL, NULL, NULL, NULL);
           if (rc)
-            return unlock_pinentry (rc);
+            return unlock_pinentry (ctrl, rc);
         }
 
       saveflag = assuan_get_flag (entry_ctx, ASSUAN_CONFIDENTIAL);
       assuan_begin_confidential (entry_ctx);
-      pinentry_status = 0;
       rc = assuan_transact (entry_ctx, "GETPIN", getpin_cb, &parm,
                             inq_quality, entry_ctx,
-                            pinentry_status_cb, &pinentry_status);
+                            pinentry_status_cb, &pininfo->status);
       assuan_set_flag (entry_ctx, ASSUAN_CONFIDENTIAL, saveflag);
       /* Most pinentries out in the wild return the old Assuan error code
          for canceled which gets translated to an assuan Cancel error and
@@ -1112,7 +1130,7 @@ agent_askpin (ctrl_t ctrl,
 
       /* Change error code in case the window close button was clicked
          to cancel the operation.  */
-      if ((pinentry_status & PINENTRY_STATUS_CLOSE_BUTTON)
+      if ((pininfo->status & PINENTRY_STATUS_CLOSE_BUTTON)
 	  && gpg_err_code (rc) == GPG_ERR_CANCELED)
         rc = gpg_err_make (gpg_err_source (rc), GPG_ERR_FULLY_CANCELED);
 
@@ -1120,7 +1138,7 @@ agent_askpin (ctrl_t ctrl,
         errtext = is_pin? L_("PIN too long")
                         : L_("Passphrase too long");
       else if (rc)
-        return unlock_pinentry (rc);
+        return unlock_pinentry (ctrl, rc);
 
       if (!errtext && pininfo->min_digits)
         {
@@ -1139,31 +1157,38 @@ agent_askpin (ctrl_t ctrl,
           /* More checks by utilizing the optional callback. */
           pininfo->cb_errtext = NULL;
           rc = pininfo->check_cb (pininfo);
-          if (gpg_err_code (rc) == GPG_ERR_BAD_PASSPHRASE
-              && pininfo->cb_errtext)
-            errtext = pininfo->cb_errtext;
-          else if (gpg_err_code (rc) == GPG_ERR_BAD_PASSPHRASE
-                   || gpg_err_code (rc) == GPG_ERR_BAD_PIN)
-            errtext = (is_pin? L_("Bad PIN") : L_("Bad Passphrase"));
+          /* When pinentry cache causes an error, return now.  */
+          if (rc
+              && (pininfo->status & PINENTRY_STATUS_PASSWORD_FROM_CACHE))
+            return unlock_pinentry (ctrl, rc);
+
+          if (gpg_err_code (rc) == GPG_ERR_BAD_PASSPHRASE)
+            {
+              if (pininfo->cb_errtext)
+                errtext = pininfo->cb_errtext;
+              else if (gpg_err_code (rc) == GPG_ERR_BAD_PASSPHRASE
+                       || gpg_err_code (rc) == GPG_ERR_BAD_PIN)
+                errtext = (is_pin? L_("Bad PIN") : L_("Bad Passphrase"));
+            }
           else if (rc)
-            return unlock_pinentry (rc);
+            return unlock_pinentry (ctrl, rc);
         }
 
       if (!errtext)
         {
           if (pininfo->with_repeat
-	      && (pinentry_status & PINENTRY_STATUS_PIN_REPEATED))
+              && (pininfo->status & PINENTRY_STATUS_PIN_REPEATED))
             pininfo->repeat_okay = 1;
-          return unlock_pinentry (0); /* okay, got a PIN or passphrase */
+          return unlock_pinentry (ctrl, 0); /* okay, got a PIN or passphrase */
         }
 
-      if ((pinentry_status & PINENTRY_STATUS_PASSWORD_FROM_CACHE))
+      if ((pininfo->status & PINENTRY_STATUS_PASSWORD_FROM_CACHE))
 	/* The password was read from the cache.  Don't count this
 	   against the retry count.  */
 	pininfo->failed_tries --;
     }
 
-  return unlock_pinentry (gpg_error (pininfo->min_digits? GPG_ERR_BAD_PIN
+  return unlock_pinentry (ctrl, gpg_error (pininfo->min_digits? GPG_ERR_BAD_PIN
                           : GPG_ERR_BAD_PASSPHRASE));
 }
 
@@ -1229,7 +1254,7 @@ agent_get_passphrase (ctrl_t ctrl,
   rc = assuan_transact (entry_ctx, line,
 			NULL, NULL, NULL, NULL, NULL, NULL);
   if (rc && gpg_err_code (rc) != GPG_ERR_ASS_UNKNOWN_CMD)
-    return unlock_pinentry (rc);
+    return unlock_pinentry (ctrl, rc);
 
 
   if (desc)
@@ -1238,18 +1263,18 @@ agent_get_passphrase (ctrl_t ctrl,
     snprintf (line, DIM(line), "RESET");
   rc = assuan_transact (entry_ctx, line, NULL, NULL, NULL, NULL, NULL, NULL);
   if (rc)
-    return unlock_pinentry (rc);
+    return unlock_pinentry (ctrl, rc);
 
   snprintf (line, DIM(line), "SETPROMPT %s", prompt);
   rc = assuan_transact (entry_ctx, line, NULL, NULL, NULL, NULL, NULL, NULL);
   if (rc)
-    return unlock_pinentry (rc);
+    return unlock_pinentry (ctrl, rc);
 
   if (with_qualitybar && opt.min_passphrase_len)
     {
       rc = setup_qualitybar (ctrl);
       if (rc)
-        return unlock_pinentry (rc);
+        return unlock_pinentry (ctrl, rc);
     }
 
   if (errtext)
@@ -1257,14 +1282,14 @@ agent_get_passphrase (ctrl_t ctrl,
       snprintf (line, DIM(line), "SETERROR %s", errtext);
       rc = assuan_transact (entry_ctx, line, NULL, NULL, NULL, NULL, NULL, NULL);
       if (rc)
-        return unlock_pinentry (rc);
+        return unlock_pinentry (ctrl, rc);
     }
 
   memset (&parm, 0, sizeof parm);
   parm.size = ASSUAN_LINELENGTH/2 - 5;
   parm.buffer = gcry_malloc_secure (parm.size+10);
   if (!parm.buffer)
-    return unlock_pinentry (out_of_core ());
+    return unlock_pinentry (ctrl, out_of_core ());
 
   saveflag = assuan_get_flag (entry_ctx, ASSUAN_CONFIDENTIAL);
   assuan_begin_confidential (entry_ctx);
@@ -1288,7 +1313,7 @@ agent_get_passphrase (ctrl_t ctrl,
     xfree (parm.buffer);
   else
     *retpass = parm.buffer;
-  return unlock_pinentry (rc);
+  return unlock_pinentry (ctrl, rc);
 }
 
 
@@ -1332,7 +1357,7 @@ agent_get_confirmation (ctrl_t ctrl,
     rc = gpg_err_make (gpg_err_source (rc), GPG_ERR_CANCELED);
 
   if (rc)
-    return unlock_pinentry (rc);
+    return unlock_pinentry (ctrl, rc);
 
   if (ok)
     {
@@ -1340,7 +1365,7 @@ agent_get_confirmation (ctrl_t ctrl,
       rc = assuan_transact (entry_ctx,
                             line, NULL, NULL, NULL, NULL, NULL, NULL);
       if (rc)
-        return unlock_pinentry (rc);
+        return unlock_pinentry (ctrl, rc);
     }
   if (notok)
     {
@@ -1363,7 +1388,7 @@ agent_get_confirmation (ctrl_t ctrl,
                                 NULL, NULL, NULL, NULL, NULL, NULL);
 	}
       if (rc)
-        return unlock_pinentry (rc);
+        return unlock_pinentry (ctrl, rc);
     }
 
   rc = assuan_transact (entry_ctx, "CONFIRM",
@@ -1371,7 +1396,7 @@ agent_get_confirmation (ctrl_t ctrl,
   if (rc && gpg_err_source (rc) && gpg_err_code (rc) == GPG_ERR_ASS_CANCELED)
     rc = gpg_err_make (gpg_err_source (rc), GPG_ERR_CANCELED);
 
-  return unlock_pinentry (rc);
+  return unlock_pinentry (ctrl, rc);
 }
 
 
@@ -1405,7 +1430,7 @@ agent_show_message (ctrl_t ctrl, const char *desc, const char *ok_btn)
     rc = gpg_err_make (gpg_err_source (rc), GPG_ERR_CANCELED);
 
   if (rc)
-    return unlock_pinentry (rc);
+    return unlock_pinentry (ctrl, rc);
 
   if (ok_btn)
     {
@@ -1413,7 +1438,7 @@ agent_show_message (ctrl_t ctrl, const char *desc, const char *ok_btn)
       rc = assuan_transact (entry_ctx, line, NULL, NULL, NULL,
                             NULL, NULL, NULL);
       if (rc)
-        return unlock_pinentry (rc);
+        return unlock_pinentry (ctrl, rc);
     }
 
   rc = assuan_transact (entry_ctx, "CONFIRM --one-button", NULL, NULL, NULL,
@@ -1421,7 +1446,7 @@ agent_show_message (ctrl_t ctrl, const char *desc, const char *ok_btn)
   if (rc && gpg_err_source (rc) && gpg_err_code (rc) == GPG_ERR_ASS_CANCELED)
     rc = gpg_err_make (gpg_err_source (rc), GPG_ERR_CANCELED);
 
-  return unlock_pinentry (rc);
+  return unlock_pinentry (ctrl, rc);
 }
 
 
@@ -1469,19 +1494,19 @@ agent_popup_message_start (ctrl_t ctrl, const char *desc, const char *ok_btn)
     snprintf (line, DIM(line), "RESET");
   rc = assuan_transact (entry_ctx, line, NULL, NULL, NULL, NULL, NULL, NULL);
   if (rc)
-    return unlock_pinentry (rc);
+    return unlock_pinentry (ctrl, rc);
 
   if (ok_btn)
     {
       snprintf (line, DIM(line), "SETOK %s", ok_btn);
       rc = assuan_transact (entry_ctx, line, NULL,NULL,NULL,NULL,NULL,NULL);
       if (rc)
-        return unlock_pinentry (rc);
+        return unlock_pinentry (ctrl, rc);
     }
 
   err = npth_attr_init (&tattr);
   if (err)
-    return unlock_pinentry (gpg_error_from_errno (err));
+    return unlock_pinentry (ctrl, gpg_error_from_errno (err));
   npth_attr_setdetachstate (&tattr, NPTH_CREATE_JOINABLE);
 
   popup_finished = 0;
@@ -1492,7 +1517,7 @@ agent_popup_message_start (ctrl_t ctrl, const char *desc, const char *ok_btn)
       rc = gpg_error_from_errno (err);
       log_error ("error spawning popup message handler: %s\n",
                  strerror (err) );
-      return unlock_pinentry (rc);
+      return unlock_pinentry (ctrl, rc);
     }
   npth_setname_np (popup_tid, "popup-message");
 
@@ -1551,10 +1576,9 @@ agent_popup_message_stop (ctrl_t ctrl)
   /* Thread IDs are opaque, but we try our best here by resetting it
      to the same content that a static global variable has.  */
   memset (&popup_tid, '\0', sizeof (popup_tid));
-  entry_owner = NULL;
 
   /* Now we can close the connection. */
-  unlock_pinentry (0);
+  unlock_pinentry (ctrl, 0);
 }
 
 int
@@ -1580,5 +1604,5 @@ agent_clear_passphrase (ctrl_t ctrl,
   rc = assuan_transact (entry_ctx, line,
 			NULL, NULL, NULL, NULL, NULL, NULL);
 
-  return unlock_pinentry (rc);
+  return unlock_pinentry (ctrl, rc);
 }
